/* this is a fragment to denoise pictures with various methods */
/* (c) andreas.romeyke@web.de */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


int getint(FILE *fd) {
  int c, i;
  char dummy[10000];

  c = getc(fd);
  while (1) {/* find next integer */

    if (c=='#')    /* if we're at a comment, read to end of line */
      fgets(dummy,9000,fd);
    if (c==EOF) {
      fprintf(stderr, "Image %s not binary PGM.\n","is");
	  exit (-1);
	}
    if (c>='0' && c<='9')
      break;   /* found what we were looking for */
    c = getc(fd);
  }

  /* we're at the start of a number, continue until we hit a non-number */
  i = 0;
  while (1) {
    i = (i*10) + (c - '0');
    c = getc(fd);
    if (c==EOF) return (i);
    if (c<'0' || c>'9') break;
  }

  return (i);
}



void get_image(char * filename,unsigned char ** in,int * x_size,int * y_size) {
	FILE  *fd;
	char header [100];
	int  tmp;
	assert(* in == NULL); /* to avoid mem error */
	if ((fd=fopen(filename,"rb")) == NULL){
		fprintf(stderr, "Can't input image %s.\n",filename);
		exit(-1);
	}
	/* read header */
	header[0]=fgetc(fd);
	header[1]=fgetc(fd);
	if (!(header[0]=='P' && header[1]=='5')) {
		fprintf(stderr,"Image %s does not have binary PGM header.\n",filename);
		exit(-1);
	}
	*x_size = getint(fd);
	*y_size = getint(fd);
	tmp = getint(fd);

	/* alloc memory to store stuff */
	* in = (unsigned char *) malloc(*x_size * *y_size * sizeof (unsigned char)); /* should be freed, if closed */
	if (* in == NULL) {
		fprintf(stderr, "could not allocate memory for output-buffer\n");
		exit(-1);
	}
	if (fread(* in,1,*x_size * *y_size,fd) == 0) {
		if (* in != NULL) {
			free(* in);
			* in=NULL;
		}
		fprintf(stderr, "Image %s is wrong size.\n",filename);
		exit(-1);
	}
	fclose(fd);
}


void put_image(char * filename,char * in,int x_size,int y_size) {
	FILE  *fd;

	if ((fd=fopen(filename,"wb")) == NULL) {
		if (in != NULL) {
			free(in);
			in=NULL;
		}
		fprintf(stderr,"Can't output image%s.\n",filename);
		exit(-1);
	}
	fprintf(fd,"P5\n");
	fprintf(fd,"%d %d\n",x_size,y_size);
	fprintf(fd,"255\n");

	if (fwrite(in,x_size*y_size,1,fd) != 1) {
		if (in != NULL) {
			free(in);
			in=NULL;
		}
		fprintf(stderr, "Can't write image %s.\n",filename);
		exit(-1);
	}
	fclose(fd);
}


void simple_autocorrel_filter(unsigned char ** in, unsigned char ** out, int * xsize, int * ysize) {
#define PIX(a,b) ( tin[((b) * maxx) + (a)])
#define OPI(a,b) ( tout[((b) * maxx) + (a)])
	unsigned char * tin;
	unsigned char * tout;
	int maxx, maxy;
	int x, y;
	int j, minj;
	double min;
#define mincorrpix 128
	double delta; /* mincorrpix entries to hold minimum about 2 lines */
	assert(*in != NULL);
	assert(*out == NULL);
	maxx= *xsize;
	maxy= *ysize;
	/* alloc memory to store stuff */
	* out = (unsigned char *) malloc(maxx * maxy *sizeof(unsigned char)); /* should be freed, if closed */
	if (* out == NULL) {
		fprintf(stderr, "could not allocate memory for output-buffer\n");
		exit(-1);
	}
	/* init with original image */
	*out = memcpy(*out, *in, maxx * maxy * sizeof(unsigned char));
	tin= *in;
	tout= *out;
	/* fill target with white pixels, to detect autocorr-jitter */
	for (y=0; y<(*ysize);y++) {
		for (x=0; x<(*xsize); x++) {
			OPI(x,y)=255;
		}
	}
	/* implement a simple autocorrelation per line */
	for (y=2; y<(*ysize)-2; y++) {
		j=0; min=256*256*256; minj=0;
		for (j= -mincorrpix/2; j< mincorrpix/2; j++) {
			delta=0;
			for (x=mincorrpix/2; x<((*xsize)-mincorrpix/2); x++) {
				delta+=
				(
					((PIX(x+j,y)-PIX(x,y-1)))*
					((PIX(x+j,y)-PIX(x,y-1)))+
					((PIX(x+j,y)-PIX(x,y+1)))*
					((PIX(x+j,y)-PIX(x,y+1)))+
					((PIX(x+j,y)-PIX(x,y-2)))*
					((PIX(x+j,y)-PIX(x,y-2)))+
					((PIX(x+j,y)-PIX(x,y+2)))*
					((PIX(x+j,y)-PIX(x,y+2)))
					);
			}
			if (min > delta) {
				min=delta;
				minj=j;
			}
		}
		for (x=mincorrpix/2; x<((*xsize)-mincorrpix/2); x++) {
//			OPI(x+minj,y)=PIX(x,y);
			OPI(x-minj,y)=PIX(x,y);
		}
	}
}


int main(int argc, char * argv[]) {
	unsigned char * in;
	unsigned char * out;
	int x;
	int y;
	if (argc !=3) {
		fprintf(stderr, "try \"%s <input filename> <output filename\" instead\n", argv[0]);
		exit(-1);
	}
	in = NULL;
	out = NULL;
	get_image(argv[1],&in, &x, &y);
	simple_autocorrel_filter(&in,&out,&x, &y);
	put_image(argv[2],out, x,y);
	return (0);
}

