/* this is a fragment to denoise pictures with various methods */
/* (c) andreas.romeyke@web.de */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>



int getint(FILE *fd) {
  int c, i;
  char dummy[10000];

  c = getc(fd);
  while (1) {/* find next integer */

    if (c=='#')    /* if we're at a comment, read to end of line */
      fgets(dummy,9000,fd);
    if (c==EOF) {
      fprintf(stderr, "Image %s not binary PGM.\n","is");
	  exit (-1);
	}
    if (c>='0' && c<='9')
      break;   /* found what we were looking for */
    c = getc(fd);
  }

  /* we're at the start of a number, continue until we hit a non-number */
  i = 0;
  while (1) {
    i = (i*10) + (c - '0');
    c = getc(fd);
    if (c==EOF) return (i);
    if (c<'0' || c>'9') break;
  }

  return (i);
}



void get_image(char * filename,unsigned char ** in,int * x_size,int * y_size) {
	FILE  *fd;
	char header [100];
	int  tmp;
	assert(* in == NULL); /* to avoid mem error */
	if ((fd=fopen(filename,"rb")) == NULL){
		fprintf(stderr, "Can't input image %s.\n",filename);
		exit(-1);
	}
	/* read header */
	header[0]=fgetc(fd);
	header[1]=fgetc(fd);
	if (!(header[0]=='P' && header[1]=='5')) {
		fprintf(stderr,"Image %s does not have binary PGM header.\n",filename);
		exit(-1);
	}
	*x_size = getint(fd);
	*y_size = getint(fd);
	tmp = getint(fd);

	/* alloc memory to store stuff */
	* in = (unsigned char *) malloc(*x_size * *y_size * sizeof (unsigned char)); /* should be freed, if closed */
	if (* in == NULL) {
		fprintf(stderr, "could not allocate memory for output-buffer\n");
		exit(-1);
	}
	if (fread(* in,1,*x_size * *y_size,fd) == 0) {
		if (* in != NULL) {
			free(* in);
			* in=NULL;
		}
		fprintf(stderr, "Image %s is wrong size.\n",filename);
		exit(-1);
	}
	fclose(fd);
}


void put_image(char * filename,char * in,int x_size,int y_size) {
	FILE  *fd;

	if ((fd=fopen(filename,"wb")) == NULL) {
		if (in != NULL) {
			free(in);
			in=NULL;
		}
		fprintf(stderr,"Can't output image%s.\n",filename);
		exit(-1);
	}
	fprintf(fd,"P5\n");
	fprintf(fd,"%d %d\n",x_size,y_size);
	fprintf(fd,"255\n");

	if (fwrite(in,x_size*y_size,1,fd) != 1) {
		if (in != NULL) {
			free(in);
			in=NULL;
		}
		fprintf(stderr, "Can't write image %s.\n",filename);
		exit(-1);
	}
	fclose(fd);
}

/* this calcs a gradienten regression on 5 pixel blocks, it is not too bad, but tends to produce JPEG-like blockartefacts. */

void simple_linregr_filter(unsigned char ** in, unsigned char ** out, int * xsize, int * ysize) {
#define PIX(a,b) ( tin[((b) * maxx) + (a)])
#define OPI(a,b) ( tout[((b) * maxx) + (a)])
	unsigned char * tin;
	unsigned char * tout;
	int maxx, maxy;
	int Y;
	int x,y;

	float a, b; /* a+bx */
	float meanx, meany;
	float sumxy, sumxx;
	float sumx, sumy;

	assert(*in != NULL);
	assert(*out == NULL);
	maxx= *xsize;
	maxy= *ysize;
	/* alloc memory to store stuff */
	* out = (unsigned char *) malloc(maxx * maxy *sizeof(unsigned char)); /* should be freed, if closed */
	if (* out == NULL) {
		fprintf(stderr, "could not allocate memory for output-buffer\n");
		exit(-1);
	}
	/* init with original image */
	*out = memcpy(*out, *in, maxx * maxy * sizeof(unsigned char));
	tin= *in;
	tout= *out;
	/* implement a simple linregr */
	for (y=2; y<maxy-8; y+=1) {
#ifndef DEMO
		for (x=2; x<maxx-8; x+=8) {
#else
		for (x=2; x<y; x+=8) {
#endif
				sumxy=
					(x)*PIX((x),y) +
					(x+1)*PIX((x+1),y)+
					(x+2)*PIX((x+2),y)+
					(x+3)*PIX((x+3),y)+
					(x+4)*PIX((x+4),y) +
					(x+5)*PIX((x+5),y)+
					(x+6)*PIX((x+6),y)+
					(x+7)*PIX((x+7),y);
				sumx=
					(x)+
					(x+1)+
					(x+2)+
					(x+3)+
					(x+4)+
					(x+5)+
					(x+6)+
					(x+7);
				sumxx=
					(x)*(x)+
					(x+1)*(x+1)+
					(x+2)*(x+2)+
					(x+3)*(x+3)+
					(x+4)*(x+4)+
					(x+5)*(x+5)+
					(x+6)*(x+6)+
					(x+7)*(x+7);
				sumy=
					PIX(x,y)+
					PIX(x+1,y)+
					PIX(x+2,y)+
					PIX(x+3,y)+
					PIX(x+4,y)+
					PIX(x+5,y)+
					PIX(x+6,y)+
					PIX(x+7,y);
				meanx=sumx/8;
				meany=sumy/8;
				b= (sumxy - 8*meanx*meany)/(sumxx-8*meanx*meanx);
				a= meany - b*meanx;

				Y=a+b*(x);
				if (Y > 255) Y=255;
				OPI(x,y)= Y;

				Y=a+b*(x+1);
				if (Y > 255) Y=255;
				OPI(x+2,y)=Y;

				Y=a+b*(x+2);
				if (Y > 255) Y=255;
				OPI(x+4,y)=Y;

				Y=a+b*(x+3);
				if (Y > 255) Y=255;
				OPI(x+6,y)=Y;

				Y=a+b*(x+4);
				if (Y > 255) Y=255;
				OPI(x+8,y)= Y;

				Y=a+b*(x+5);
				if (Y > 255) Y=255;
				OPI(x+10,y)=Y;

				Y=a+b*(x+6);
				if (Y > 255) Y=255;
				OPI(x+12,y)=Y;

				Y=a+b*(x+7);
				if (Y > 255) Y=255;
				OPI(x+14,y)=Y;


		}
	}
	/* copy back */
	for (y=2; y<maxy-2; y++) {
		for (x=2; x<maxx-2; x++) {
			PIX(x,y)=OPI(x,y);
		}
	}

	for (y=2; y<maxy-8; y+=8) {
#ifndef DEMO
		for (x=2; x<maxx-8; x+=1) {
#else
		for (x=2; x<y; x+=1) {
#endif
				sumxy=
					(y)*PIX((x),y) +
					(y+1)*PIX((x),y+1)+
					(y+2)*PIX((x),y+2)+
					(y+3)*PIX((x),y+3)+
					(y+4)*PIX((x),y+4) +
					(y+5)*PIX((x),y+5)+
					(y+6)*PIX((x),y+6)+
					(y+7)*PIX((x),y+7);
				sumx=
					(y)+
					(y+1)+
					(y+2)+
					(y+3)+
					(y+4)+
					(y+5)+
					(y+6)+
					(y+7);
				sumxx=
					(y)*(x)+
					(y+1)*(y+1)+
					(y+2)*(y+2)+
					(y+3)*(y+3)+
					(y+4)*(y+4)+
					(y+5)*(y+5)+
					(y+6)*(y+6)+
					(y+7)*(y+7);
				sumy=
					PIX(x,y)+
					PIX(x,y+1)+
					PIX(x,y+2)+
					PIX(x,y+3)+
					PIX(x,y+4)+
					PIX(x,y+5)+
					PIX(x,y+6)+
					PIX(x,y+7);
				meanx=sumx/8;
				meany=sumy/8;
				b= (sumxy - 8*meanx*meany)/(sumxx-8*meanx*meanx);
				a= meany - b*meanx;

				Y=a+b*(y+1);
				if (Y > 255) Y=255;
				OPI(x,y+1)=Y;

				Y=a+b*(y+2);
				if (Y > 255) Y=255;
				OPI(x,y+2)=Y;

				Y=a+b*(y+3);
				if (Y > 255) Y=255;
				OPI(x,y+3)=Y;

				Y=a+b*(y+4);
				if (Y > 255) Y=255;
				OPI(x,y+4)= Y;

				Y=a+b*(y+5);
				if (Y > 255) Y=255;
				OPI(x,y+5)=Y;

				Y=a+b*(y+6);
				if (Y > 255) Y=255;
				OPI(x,y+6)=Y;

				Y=a+b*(y+7);
				if (Y > 255) Y=255;
				OPI(x,y+7)=Y;
		}
	}

}



int main(int argc, char * argv[]) {
	unsigned char * in;
	unsigned char * out;
	int x;
	int y;
	if (argc !=3) {
		fprintf(stderr, "try \"%s <input filename> <output filename\" instead\n", argv[0]);
		exit(-1);
	}
	in = NULL;
	out = NULL;
	get_image(argv[1],&in, &x, &y);
	simple_linregr_filter(&in,&out,&x, &y);
	put_image(argv[2],out, x,y);
	return (0);
}

