/* this is a fragment to denoise pictures with various methods */
/* (c) andreas.romeyke@web.de */

/* TODO: Bestimme Regressionsgeraden ua. mit gleitendem Mittelwertfenster
   so, da a) zwei Geraden sich in einem Punkt im Intervall 0<=y<=255 schneiden
   */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


int getint(FILE *fd) {
  int c, i;
  char dummy[10000];

  c = getc(fd);
  while (1) {/* find next integer */

    if (c=='#')    /* if we're at a comment, read to end of line */
      fgets(dummy,9000,fd);
    if (c==EOF) {
      fprintf(stderr, "Image %s not binary PGM.\n","is");
	  exit (-1);
	}
    if (c>='0' && c<='9')
      break;   /* found what we were looking for */
    c = getc(fd);
  }

  /* we're at the start of a number, continue until we hit a non-number */
  i = 0;
  while (1) {
    i = (i*10) + (c - '0');
    c = getc(fd);
    if (c==EOF) return (i);
    if (c<'0' || c>'9') break;
  }

  return (i);
}



void get_image(char * filename,unsigned char ** in,int * x_size,int * y_size) {
	FILE  *fd;
	char header [100];
	int  tmp;
	assert(* in == NULL); /* to avoid mem error */
	if ((fd=fopen(filename,"rb")) == NULL){
		fprintf(stderr, "Can't input image %s.\n",filename);
		exit(-1);
	}
	/* read header */
	header[0]=fgetc(fd);
	header[1]=fgetc(fd);
	if (!(header[0]=='P' && header[1]=='5')) {
		fprintf(stderr,"Image %s does not have binary PGM header.\n",filename);
		exit(-1);
	}
	*x_size = getint(fd);
	*y_size = getint(fd);
	tmp = getint(fd);

	/* alloc memory to store stuff */
	* in = (unsigned char *) malloc(*x_size * *y_size * sizeof (unsigned char)); /* should be freed, if closed */
	if (* in == NULL) {
		fprintf(stderr, "could not allocate memory for output-buffer\n");
		exit(-1);
	}
	if (fread(* in,1,*x_size * *y_size,fd) == 0) {
		if (* in != NULL) {
			free(* in);
			* in=NULL;
		}
		fprintf(stderr, "Image %s is wrong size.\n",filename);
		exit(-1);
	}
	fclose(fd);
}


void put_image(char * filename,char * in,int x_size,int y_size) {
	FILE  *fd;

	if ((fd=fopen(filename,"wb")) == NULL) {
		if (in != NULL) {
			free(in);
			in=NULL;
		}
		fprintf(stderr,"Can't output image%s.\n",filename);
		exit(-1);
	}
	fprintf(fd,"P5\n");
	fprintf(fd,"%d %d\n",x_size,y_size);
	fprintf(fd,"255\n");

	if (fwrite(in,x_size*y_size,1,fd) != 1) {
		if (in != NULL) {
			free(in);
			in=NULL;
		}
		fprintf(stderr, "Can't write image %s.\n",filename);
		exit(-1);
	}
	fclose(fd);
}

/* delta calculation, returns 0 if no change, 1 if change */
inline int delta(int prev, int act, int j) {
	if (((prev-act)< 9)&((prev-act)> -9)) { /* ist gleich geblieben, 9 meint: 1, -9 meint -1 */
		return(0);
	} else return(1);
}

inline int median(int a, int b, int c) {
	if (a<=b && b<=c) return b;
	if (b<=c && c<=a) return c;
	if (c<=a && a<=b) return a;
	if (c<=b && b<=a) return b;
	if (a<=c && c<=b) return c;
	if (b<=a && a<=c) return a;
	return 0;
}


/* this calcs a gradienten regression on 5 pixel blocks, it is not too bad, but tends to produce JPEG-like blockartefacts. */
void simple_linregr_adapt_filter(unsigned char ** in, unsigned char ** out, int * xsize, int * ysize) {
#define PIX(a,b) ( tin[((b) * maxx) + (a)])
#define OPI(a,b) ( tout[((b) * maxx) + (a)])
/* SMOOTHNESS defines Threshold for impuls-detection, should be between 0.01 and 0.06, a good value is 0.05 */
#define SMOOTHNESS (0.05)
/* MAXSTEPWIDE defines maximum count of values which should be in a linregression, default is 8 */
#define MAXSTEPWIDE (8)
	unsigned char * tin;
	unsigned char * tout;
	int maxx, maxy;
	int i,Y;
	int x,y;
	int j;
	float a, b; /* a+bx */
	float meanx, meany;
	float sumxy, sumxx;
	float sumx, sumy;
	long int prevmean;
	long int actmean;
	assert(*in != NULL);
	assert(*out == NULL);
	maxx= *xsize;
	maxy= *ysize;
	/* alloc memory to store stuff */
	* out = (unsigned char *) malloc(maxx * maxy *sizeof(unsigned char)); /* should be freed, if closed */
	if (* out == NULL) {
		fprintf(stderr, "could not allocate memory for output-buffer\n");
		exit(-1);
	}
	/* init with original image */
	*out = memcpy(*out, *in, maxx * maxy * sizeof(unsigned char));
	tin= *in;
	tout= *out;
	/* implement a simple linregr */
#ifndef BLXA
	for (y=4; y<maxy-4; y++) {
#ifndef DEMO
		for (x=4; x<maxx-4; ) {
#else
		for (x=4; x<y; ) {
#endif
			/* search endwindow */
			/* calc prevmean */
			prevmean=
			(
				PIX(x+1,y)+
				PIX(x+2,y)+
				PIX(x+3,y)+
				PIX(x+4,y)+
				PIX(x-4,y)+
				PIX(x-3,y)+
				PIX(x-2,y)+
				PIX(x-1,y)+
				PIX(x-0,y)
			);

			for (j=1; j<maxx-4-x;j++) { /* to slide window forwards */
				if (j>MAXSTEPWIDE) break;
				/* calc actmean */
				actmean=
				(
					PIX(x+0+j,y)+
					PIX(x+1+j,y)+
					PIX(x+2+j,y)+
					PIX(x+3+j,y)+
					PIX(x+4+j,y)+
					PIX(x-1+j,y)+
					PIX(x-2+j,y)+
					PIX(x-3+j,y)+
					PIX(x-4+j,y)
					);
					/* filter noisy peaks out */
					if (abs(PIX(x+j,y)*9-(actmean))>(0.05*actmean)) {
						/* substitute it with 3x3 median */
						//actmean=actmean*9;
						actmean=actmean-PIX(x+j,y);
						PIX(x+j,y)=median(PIX(x+j-1,y),PIX(x+j,y),PIX(x+j+1,y));
						actmean=actmean+PIX(x+j,y);
						//actmean=actmean/9;
					}

					if (delta(prevmean,actmean,j)==0) {/* same thing as before, slide window forwards */
					;
				}
				else {/* new situation, mark window position, recalc regression and forward startpoint to window position */
					break; /* break out the for loop, to do the things described above */
				}
			}
			sumxy=0;
			sumx=0;
			sumxx=0;
			sumy=0;
			meanx=0;
			meany=0;
			for (i=x; i<=(x+j); i++) {
				sumxy+=
					(i)*PIX(i,y);
				sumx+=(i);
				sumxx+=(i)*(i);
				sumy+=
					PIX(i,y);
			}
			meanx=sumx/j;
			meany=sumy/j;
			b= (sumxy - j*meanx*meany)/(sumxx-j*meanx*meanx);
			a= meany - b*meanx;
			for (i=x; i<=(x+j); i++) {
				Y=a+b*(i);
				if (Y > 255) Y=255;
				if (Y < 0) Y=0;
				OPI(i,y)= Y;
			}
			x+=j; /* set actpos to windowpos */
		}
	}


	/* copy back
	for (y=2; y<maxy-2; y++) {
		for (x=2; x<maxx-2; x++) {
			PIX(x,y)=OPI(x,y);
		}
	}

	*/

#endif

	for (x=4; x<maxx-4; x++) {
		for (y=4; y<maxy-4; ) {
			/* search endwindow */
			/* calc prevmean */
			prevmean=
			(
				PIX(x,y+1)+
				PIX(x,y+2)+
				PIX(x,y+3)+
				PIX(x,y+4)+
				PIX(x,y-4)+
				PIX(x,y-3)+
				PIX(x,y-2)+
				PIX(x,y-1)+
				PIX(x,y)
			);

			for (j=1; j<maxy-4-y;j++) { /* to slide window forwards */
				if (j>MAXSTEPWIDE) break;
				/* calc actmean */
				actmean=
				(
					PIX(x,y+0+j)+
					PIX(x,y+1+j)+
					PIX(x,y+2+j)+
					PIX(x,y+3+j)+
					PIX(x,y+4+j)+
					PIX(x,y-1+j)+
					PIX(x,y-2+j)+
					PIX(x,y-3+j)+
					PIX(x,y-4+j)
					);
					/* filter noisy peaks out */
					if (abs(PIX(x+j,y)*9-(actmean))>(0.05*actmean)) {
//					if (abs(PIX(x,y+j)*9-actmean)>(7)) {
						/* substitute it with 3x3 median */
						//actmean=actmean*9;
						actmean=actmean-PIX(x,y+j);
						PIX(x,y+j)=median(PIX(x,y+j-1),PIX(x,y+j),PIX(x,y+j+1));
						actmean=actmean+PIX(x,y+j);
						//actmean=actmean/9;
					}

					if (delta(prevmean,actmean,j)==0) {/* same thing as before, slide window forwards */
					;
				}
				else {/* new situation, mark window position, recalc regression and forward startpoint to window position */
					break; /* break out the for loop, to do the things described above */
				}
			}
			sumxy=0;
			sumx=0;
			sumxx=0;
			sumy=0;
			meanx=0;
			meany=0;
			for (i=y; i<=(y+j); i++) {
				sumxy+=
					(i)*PIX(x,i);
				sumx+=(i);
				sumxx+=(i)*(i);
				sumy+=
					PIX(x,i);
			}
			meanx=sumx/j;
			meany=sumy/j;
			b= (sumxy - j*meanx*meany)/(sumxx-j*meanx*meanx);
			a= meany - b*meanx;
			for (i=y; i<=(y+j); i++) {
				Y=a+b*(i);
				if (Y > 255) Y=255;
				if (Y < 0) Y=0;
				OPI(x,i)=((Y)+PIX(x,i))/2;
			}

			y+=j; /* set actpos to windowpos */
		}
	}

}



int main(int argc, char * argv[]) {
	unsigned char * in;
	unsigned char * out;
	int x;
	int y;
	if (argc !=3) {
		fprintf(stderr, "try \"%s <input filename> <output filename\" instead\n", argv[0]);
		exit(-1);
	}
	in = NULL;
	out = NULL;
	get_image(argv[1],&in, &x, &y);
	simple_linregr_adapt_filter(&in,&out,&x, &y);
	put_image(argv[2],out, x,y);
	return (0);
}

