/* this is a fragment to denoise pictures with various methods */
/* (c) andreas.romeyke@web.de */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


int getint(FILE *fd) {
  int c, i;
  char dummy[10000];

  c = getc(fd);
  while (1) {/* find next integer */

    if (c=='#')    /* if we're at a comment, read to end of line */
      fgets(dummy,9000,fd);
    if (c==EOF) {
      fprintf(stderr, "Image %s not binary PGM.\n","is");
	  exit (-1);
	}
    if (c>='0' && c<='9')
      break;   /* found what we were looking for */
    c = getc(fd);
  }

  /* we're at the start of a number, continue until we hit a non-number */
  i = 0;
  while (1) {
    i = (i*10) + (c - '0');
    c = getc(fd);
    if (c==EOF) return (i);
    if (c<'0' || c>'9') break;
  }

  return (i);
}



void get_image(char * filename,unsigned char ** in,int * x_size,int * y_size) {
	FILE  *fd;
	char header [100];
	int  tmp;
	assert(* in == NULL); /* to avoid mem error */
	if ((fd=fopen(filename,"rb")) == NULL){
		fprintf(stderr, "Can't input image %s.\n",filename);
		exit(-1);
	}
	/* read header */
	header[0]=fgetc(fd);
	header[1]=fgetc(fd);
	if (!(header[0]=='P' && header[1]=='5')) {
		fprintf(stderr,"Image %s does not have binary PGM header.\n",filename);
		exit(-1);
	}
	*x_size = getint(fd);
	*y_size = getint(fd);
	tmp = getint(fd);

	/* alloc memory to store stuff */
	* in = (unsigned char *) malloc(*x_size * *y_size * sizeof (unsigned char)); /* should be freed, if closed */
	if (* in == NULL) {
		fprintf(stderr, "could not allocate memory for output-buffer\n");
		exit(-1);
	}
	if (fread(* in,1,*x_size * *y_size,fd) == 0) {
		if (* in != NULL) {
			free(* in);
			* in=NULL;
		}
		fprintf(stderr, "Image %s is wrong size.\n",filename);
		exit(-1);
	}
	fclose(fd);
}


void put_image(char * filename,char * in,int x_size,int y_size) {
	FILE  *fd;

	if ((fd=fopen(filename,"wb")) == NULL) {
		if (in != NULL) {
			free(in);
			in=NULL;
		}
		fprintf(stderr,"Can't output image%s.\n",filename);
		exit(-1);
	}
	fprintf(fd,"P5\n");
	fprintf(fd,"%d %d\n",x_size,y_size);
	fprintf(fd,"255\n");

	if (fwrite(in,x_size*y_size,1,fd) != 1) {
		if (in != NULL) {
			free(in);
			in=NULL;
		}
		fprintf(stderr, "Can't write image %s.\n",filename);
		exit(-1);
	}
	fclose(fd);
}


void simple_median_filter(unsigned char ** in, unsigned char ** out, int * xsize, int * ysize) {
#define PIX(a,b) ( tin[((b) * maxx) + (a)])
#define OPI(a,b) ( tout[((b) * maxx) + (a)])
	unsigned char * tin;
	unsigned char * tout;
	int maxx, maxy;
	int x, y;
	int i, j, min,Y;
	int temp[25]; /* 5x5 matrix */
	assert(*in != NULL);
	assert(*out == NULL);
	maxx= *xsize;
	maxy= *ysize;
	/* alloc memory to store stuff */
	* out = (unsigned char *) malloc(maxx * maxy *sizeof(unsigned char)); /* should be freed, if closed */
	if (* out == NULL) {
		fprintf(stderr, "could not allocate memory for output-buffer\n");
		exit(-1);
	}
	/* init with original image */
	*out = memcpy(*out, *in, maxx * maxy * sizeof(unsigned char));
	tin= *in;
	tout= *out;
	/* implement a simple median */
	for (y=2; y<maxy-2; y++) {
#ifndef DEMO
		for (x=2; x<maxx-2; x++) {
#else
		for (x=2; x<y; x++) {
#endif
			temp[0]=PIX(x-2,y-2);
			temp[1]=PIX(x-1,y-2);
			temp[2]=PIX(x-0,y-2);
			temp[3]=PIX(x+1,y-2);
			temp[4]=PIX(x+2,y-2);
			temp[5]=PIX(x-2,y-1);
			temp[6]=PIX(x-1,y-1);
			temp[7]=PIX(x-0,y-1);
			temp[8]=PIX(x+1,y-1);
			temp[9]=PIX(x+2,y-1);
			temp[10]=PIX(x-2,y);
			temp[11]=PIX(x-1,y);
			temp[12]=PIX(x-0,y);
			temp[13]=PIX(x+1,y);
			temp[14]=PIX(x+2,y);
			temp[15]=PIX(x-2,y+1);
			temp[16]=PIX(x-1,y+1);
			temp[17]=PIX(x-0,y+1);
			temp[18]=PIX(x+1,y+1);
			temp[19]=PIX(x+2,y+1);
			temp[20]=PIX(x-2,y+2);
			temp[21]=PIX(x-1,y+2);
			temp[22]=PIX(x-0,y+2);
			temp[23]=PIX(x+1,y+2);
			temp[24]=PIX(x+2,y+2);
			Y=0;
			for (i=0; i<24-1; i++) { /* selection sort temp to get the median */
				min=i;
				for (j=i+1; j<24; j++) {
					if (temp[j]<temp[min]) {
						min=j;
						Y=temp[min]; temp[min]=temp[i]; temp[i]=Y;
					}
				}
			}
			OPI(x,y)=temp[4];
		}
	}
}



int main(int argc, char * argv[]) {
	unsigned char * in;
	unsigned char * out;
	int x;
	int y;
	if (argc !=3) {
		fprintf(stderr, "try \"%s <input filename> <output filename\" instead\n", argv[0]);
		exit(-1);
	}
	in = NULL;
	out = NULL;
	get_image(argv[1],&in, &x, &y);
	simple_median_filter(&in,&out,&x, &y);
	put_image(argv[2],out, x,y);
	return (0);
}

